@@ -64,6 +64,12 @@
 //! let luvs = rgb_bytes_to_luvs(&rgbs);
 //! ```
 //!
+//! # Features
+//!
+//! The crate defines an `approx` feature.  If enabled, approximate equality as
+//! defined by [`approx` crate](https://crates.io/crates/approx) will be
+//! implemented for the `Luv` and `LCh` types.
+//!
 //! # Other crates
 //!
 //! The design — and to some degree code — of this crate has been based on the
@@ -73,27 +79,56 @@
 //! For conversion between sRGB and XYZ colour spaces this crate relies on the
 //! [`srgb` crate](https://crates.io/crates/srgb).
 
-#[cfg(test)]
+#[cfg(any(test, feature = "approx"))]
 mod approx_impl;
 
 /// Struct representing a color in CIALuv, a.k.a. L\*u\*v\*, color space
-#[derive(Debug, PartialEq, Copy, Clone, Default)]
+#[derive(Debug, Copy, Clone, Default)]
 pub struct Luv {
+    /// The L\* value (achromatic luminance) of the colour in 0–100 range.
     pub l: f32,
+    /// The u\* value of the colour.
+    ///
+    /// Together with v\* value, it defines chromaticity of the colour.  The u\*
+    /// coordinate represents colour’s position on red-green axis with negative
+    /// values indicating more red and positive more green colour.  Typical
+    /// values are in -100–100 range (but exact range for ‘valid’ colours
+    /// depends on luminance and v\* value).
     pub u: f32,
+    /// The u\* value of the colour.
+    ///
+    /// Together with u\* value, it defines chromaticity of the colour.  The v\*
+    /// coordinate represents colour’s position on blue-yellow axis with
+    /// negative values indicating more blue and positive more yellow colour.
+    /// Typical values are in -100–100 range (but exact range for ‘valid’
+    /// colours depends on luminance and u\* value).
     pub v: f32,
 }
 
 /// Struct representing a color in cylindrical CIELCh(uv) color space
-#[derive(Debug, PartialEq, Copy, Clone, Default)]
+#[derive(Debug, Copy, Clone, Default)]
 pub struct LCh {
+    /// The L\* value (achromatic luminance) of the colour in 0–100 range.
+    ///
+    /// This is the same value as in the [`Luv`] object.
     pub l: f32,
+    /// The C\*_uv value (chroma) of the colour.
+    ///
+    /// Together with h_uv, it defines chromaticity of the colour.  The typical
+    /// values of the coordinate go from zero up to around 150 (but exact range
+    /// for ‘valid’ colours depends on luminance and hue).  Zero represents
+    /// shade of grey.
     pub c: f32,
+    /// The h_uv value (hue) of the colour measured in radians.
+    ///
+    /// Together with C\*_uv, it defines chromaticity of the colour.  The value
+    /// represents an angle thus it wraps around τ.  Typically, the value will
+    /// be in the -π–π range.  The value is undefined if C\*_uv is zero.
     pub h: f32,
 }
 
 
-// κ and ε parameters used in conversion between XYZ and La*b*.  See
+// κ and ε parameters used in conversion between XYZ and L*u*v*.  See
 // http://www.brucelindbloom.com/LContinuity.html for explanation as to why
 // those are different values than those provided by CIE standard.
 const KAPPA: f32 = 24389.0 / 27.0;
@@ -101,8 +136,23 @@ const ONE_OVER_KAPPA: f32 = 27.0 / 24389.0;
 const EPSILON: f32 = 216.0 / 24389.0;
 const KAPPA_EPSILON: f32 = /* κ * ε = 216 / 27 = 8 */ 8.0;
 
-const WHITE_U_PRIME: f32 = srgb::xyz::D65_uv[0];
-const WHITE_V_PRIME: f32 = srgb::xyz::D65_uv[1];
+use srgb::xyz::D65_XYZ;
+const WHITE_U_PRIME: f32 =
+    4.0 * D65_XYZ[0] / (D65_XYZ[0] + 15.0 * D65_XYZ[1] + 3.0 * D65_XYZ[2]);
+const WHITE_V_PRIME: f32 =
+    9.0 * D65_XYZ[1] / (D65_XYZ[0] + 15.0 * D65_XYZ[1] + 3.0 * D65_XYZ[2]);
+
+
+#[cfg(any(target_feature = "fma", test))]
+fn mul_add(multiplier: f32, multiplicand: f32, addend: f32) -> f32 {
+    multiplier.mul_add(multiplicand, addend)
+}
+
+#[cfg(not(any(target_feature = "fma", test)))]
+fn mul_add(multiplier: f32, multiplicand: f32, addend: f32) -> f32 {
+    multiplier * multiplicand + addend
+}
+
 
 fn luv_from_xyz(xyz: [f32; 3]) -> Luv {
     let [x, y, z] = xyz;
@@ -112,13 +162,13 @@ fn luv_from_xyz(xyz: [f32; 3]) -> Luv {
     } else if y <= EPSILON {
         KAPPA * y
     } else {
-        y.powf(1.0 / 3.0).mul_add(116.0, -16.0)
+        mul_add(y.powf(1.0 / 3.0), 116.0, -16.0)
     };
 
-    let d = y.mul_add(15.0, z.mul_add(3.0, x));
+    let d = mul_add(y, 15.0, mul_add(z, 3.0, x));
     let ll = 13.0 * l;
-    let u = ll * (x / d).mul_add(4.0, -WHITE_U_PRIME);
-    let v = ll * (y / d).mul_add(9.0, -WHITE_V_PRIME);
+    let u = ll * mul_add(x / d, 4.0, -WHITE_U_PRIME);
+    let v = ll * mul_add(y / d, 9.0, -WHITE_V_PRIME);
 
     Luv { l, u, v }
 }
@@ -137,9 +187,9 @@ fn xyz_from_luv(luv: &Luv) -> [f32; 3] {
         luv.l * ONE_OVER_KAPPA
     };
 
-    let a = 0.75 * u_prime / v_prime; /* 3/4 * u' * v' */
-    let x = y * 3.0 * a;
-    let z = y * (3.0 / v_prime - a - 5.0);
+    let a = 0.75 * y * u_prime / v_prime;
+    let x = 3.0 * a;
+    let z = y * (3.0 - 5.0 * v_prime) / v_prime - a;
 
     [x, y, z]
 }
@@ -152,9 +202,9 @@ fn xyz_from_luv(luv: &Luv) -> [f32; 3] {
 /// let rgbs = &[[255u8, 0, 0], [255, 0, 255], [0, 255, 255]];
 /// let luvs = luv::rgbs_to_luvs(rgbs);
 /// assert_eq!(vec![
-///     luv::Luv { l: 53.238235, u: 175.01141, v: 37.758636 },
-///     luv::Luv { l: 60.32269, u: 84.06383, v: -108.690346 },
-///     luv::Luv { l: 91.11428, u: -70.46933, v: -15.2037325 },
+///     luv::Luv { l: 53.238235, u: 175.01141, v: 37.75865 },
+///     luv::Luv { l: 60.322693, u: 84.063835, v: -108.69035 },
+///     luv::Luv { l: 91.11428, u: -70.46933, v: -15.203715 },
 /// ], luvs);
 /// ```
 #[inline]
@@ -170,9 +220,9 @@ pub fn rgbs_to_luvs(rgbs: &[[u8; 3]]) -> Vec<Luv> {
 /// let rgbs = &[255u8, 0, 0, 255, 0, 255, 0, 255, 255];
 /// let luvs = luv::rgb_bytes_to_luvs(rgbs);
 /// assert_eq!(vec![
-///     luv::Luv { l: 53.238235, u: 175.01141, v: 37.758636 },
-///     luv::Luv { l: 60.32269, u: 84.06383, v: -108.690346 },
-///     luv::Luv { l: 91.11428, u: -70.46933, v: -15.2037325 },
+///     luv::Luv { l: 53.238235, u: 175.01141, v: 37.75865 },
+///     luv::Luv { l: 60.322693, u: 84.063835, v: -108.69035 },
+///     luv::Luv { l: 91.11428, u: -70.46933, v: -15.203715 },
 /// ], luvs);
 /// ```
 pub fn rgb_bytes_to_luvs(bytes: &[u8]) -> Vec<Luv> {
@@ -236,7 +286,7 @@ impl Luv {
     ///
     /// ```
     /// let luv = luv::Luv::from_rgb(&[240, 33, 95]);
-    /// assert_eq!(luv::Luv { l: 52.334686, u: 138.98636, v: 7.8476834 }, luv);
+    /// assert_eq!(luv::Luv { l: 52.334686, u: 138.98636, v: 7.8476787 }, luv);
     /// ```
     pub fn from_rgb(rgb: &[u8; 3]) -> Self {
         luv_from_xyz(srgb::xyz_from_u8(*rgb))
@@ -257,7 +307,7 @@ impl Luv {
     ///
     /// ```
     /// let luv = luv::Luv::from_rgba(&[240, 33, 95, 255]);
-    /// assert_eq!(luv::Luv { l: 52.334686, u: 138.98636, v: 7.8476834 }, luv);
+    /// assert_eq!(luv::Luv { l: 52.334686, u: 138.98636, v: 7.8476787 }, luv);
     /// ```
     pub fn from_rgba(rgba: &[u8; 4]) -> Self { Luv::from_rgb(subarray(rgba)) }
 
@@ -307,7 +357,7 @@ impl LCh {
     /// ```
     /// let rgb = [240, 33, 95];
     /// let lch = luv::LCh::from_rgb(&rgb);
-    /// assert_eq!(luv::LCh {l: 52.334686, c: 139.20773, h: 0.056403805}, lch);
+    /// assert_eq!(luv::LCh { l: 52.334686, c: 139.20773, h: 0.05640377 }, lch);
     /// assert_eq!(lch, luv::LCh::from_luv(luv::Luv::from_rgb(&rgb)));
     /// ```
     pub fn from_rgb(rgb: &[u8; 3]) -> Self {
@@ -325,7 +375,7 @@ impl LCh {
     /// ```
     /// let rgba = [240, 33, 95, 255];
     /// let lch = luv::LCh::from_rgba(&rgba);
-    /// assert_eq!(luv::LCh {l: 52.334686, c: 139.20773, h: 0.056403805}, lch);
+    /// assert_eq!(luv::LCh { l: 52.334686, c: 139.20773, h: 0.05640377 }, lch);
     /// assert_eq!(lch, luv::LCh::from_luv(luv::Luv::from_rgba(&rgba)));
     /// ```
     pub fn from_rgba(rgba: &[u8; 4]) -> Self {
@@ -398,6 +448,40 @@ impl LCh {
     }
 }
 
+
+impl std::cmp::PartialEq<Luv> for Luv {
+    /// Compares two colours ignoring chromaticity if L\* is zero.
+    fn eq(&self, other: &Self) -> bool {
+        if self.l != other.l {
+            false
+        } else if self.l == 0.0 {
+            true
+        } else {
+            self.u == other.u && self.v == other.v
+        }
+    }
+}
+
+impl std::cmp::PartialEq<LCh> for LCh {
+    /// Compares two colours ignoring chromaticity if L\* is zero and hue if C\*
+    /// is zero.  Hues which are τ apart are compared equal.
+    fn eq(&self, other: &Self) -> bool {
+        if self.l != other.l {
+            false
+        } else if self.l == 0.0 {
+            true
+        } else if self.c != other.c {
+            false
+        } else if self.c == 0.0 {
+            true
+        } else {
+            use std::f32::consts::TAU;
+            self.h.rem_euclid(TAU) == other.h.rem_euclid(TAU)
+        }
+    }
+}
+
+
 #[cfg(test)]
 mod tests {
     use super::{LCh, Luv};
@@ -431,23 +515,23 @@ mod tests {
             [255, 255, 255],
         ],
         xyz: [
-            [0.5181152, 0.36154357, 0.2829195],
-            [0.08752622, 0.045130707, 0.0041027875],
-            [0.07589042, 0.15178084, 0.025296856],
-            [0.038297836, 0.015319134, 0.20170192],
-            [0.11418825, 0.16709997, 0.22699869],
-            [0.1258241, 0.060449857, 0.2058047],
-            [0.41241077, 0.21264932, 0.019331636],
-            [0.35758457, 0.71516913, 0.11919508],
-            [0.18045376, 0.072181515, 0.95038974],
-            [0.5380384, 0.7873506, 1.0695846],
-            [0.59286475, 0.28483093, 0.96972156],
-            [0.7699955, 0.92781854, 0.1385265],
+            [0.5181153, 0.36154357, 0.2829196],
+            [0.08752622, 0.045130707, 0.004102787],
+            [0.07589042, 0.15178084, 0.025296807],
+            [0.038297836, 0.015319134, 0.20170195],
+            [0.11418824, 0.16709997, 0.22699866],
+            [0.1258241, 0.060449857, 0.20580474],
+            [0.4124108, 0.21264932, 0.019331753],
+            [0.35758454, 0.71516913, 0.11919485],
+            [0.18045378, 0.072181515, 0.9503897],
+            [0.5380384, 0.7873506, 1.0695845],
+            [0.59286475, 0.28483093, 0.9697217],
+            [0.76999545, 0.92781854, 0.13852677],
             [0.0, 0.0, 0.0],
-            [0.048729002, 0.051269446, 0.055828184],
-            [0.20171452, 0.21223073, 0.2311017],
+            [0.04872901, 0.051269446, 0.055828184],
+            [0.20171452, 0.21223073, 0.23110169],
             [0.52465874, 0.5520114, 0.6010947],
-            [0.95044917, 1.0, 1.0889173]
+            [0.95044917, 1.0, 1.0889173],
         ],
         luv: [
             Luv { l: 66.637695, u: 93.02938, v: 9.430316 },
@@ -498,10 +582,10 @@ mod tests {
 
     fn run_test_approx<T, U>(want: &[T], f: impl Fn(&U) -> T, input: &[U])
     where
-        T: PartialEq + std::fmt::Debug + approx::RelativeEq,
-        <T as approx::AbsDiffEq>::Epsilon: Clone, {
+        T: PartialEq + std::fmt::Debug + approx::RelativeEq<Epsilon = f32>,
+    {
         let actual: Vec<_> = input.iter().map(f).collect();
-        approx::assert_abs_diff_eq!(want, &actual[..]);
+        approx::assert_abs_diff_eq!(want, &actual[..], epsilon = 0.0001);
     }
 
     #[test]
@@ -638,19 +722,59 @@ mod tests {
     #[test]
     fn test_roundtrip_error() {
         let mut error: f64 = 0.0;
-        for l in 1..=10 {
-            for u in -10..=-10 {
-                for v in -10..=-10 {
+        for l in 1..=22 {
+            for u in -22..=-22 {
+                for v in -22..=-22 {
                     let src = Luv {
-                        l: l as f32 * 10.0,
-                        u: u as f32 * 10.0,
-                        v: v as f32 * 10.0,
+                        l: l as f32 / 0.22,
+                        u: u as f32 / 0.11,
+                        v: v as f32 / 0.11,
                     };
                     let dst = super::luv_from_xyz(super::xyz_from_luv(&src));
                     error = square_error(src, dst, error);
                 }
             }
         }
-        assert_eq!(5.857145879417658, error * 1e9);
+        assert_eq!(42.49181984050665, error * 1e9);
+    }
+
+    #[test]
+    #[rustfmt::skip]
+    fn test_partial_eq() {
+        use std::f32::consts::TAU;
+
+        // Chromaticity doesn’t matter if L* is zero.
+        assert_eq!(Luv { l: 0.0, u: 0.0, v: 0.0 },
+                   Luv { l: 0.0, u: 1.0, v: 0.0 });
+        assert_eq!(Luv { l: 0.0, u: 0.0, v: 0.0 },
+                   Luv { l: 0.0, u: 0.0, v: 1.0 });
+        assert_eq!(LCh { l: 0.0, c: 0.0, h: 0.0 },
+                   LCh { l: 0.0, c: 1.0, h: 0.0 });
+        assert_eq!(LCh { l: 0.0, c: 0.0, h: 0.0 },
+                   LCh { l: 0.0, c: 0.0, h: 1.0 });
+
+        // Hue doesn’t matter if C* is zero.
+        assert_eq!(LCh { l: 100.0, c: 0.0, h: 0.0 },
+                   LCh { l: 100.0, c: 0.0, h: 1.0 });
+
+        // Hues which are τ apart are eqaul.
+        assert_eq!(LCh { l: 75.0, c: 50.0, h: 1.0 },
+                   LCh { l: 75.0, c: 50.0, h: 1.0 + 2.0 * TAU });
+        assert_eq!(LCh { l: 75.0, c: 50.0, h: 1.0 },
+                   LCh { l: 75.0, c: 50.0, h: 1.0 - TAU });
+
+        // And a few non-equal test cases.
+        assert_ne!(Luv { l: 25.0, u: 100.0, v: 75.0 },
+                   Luv { l: 50.0, u: 100.0, v: 75.0 });
+        assert_ne!(Luv { l: 50.0, u: 100.0, v: 75.0 },
+                   Luv { l: 50.0, u:  50.0, v: 75.0 });
+        assert_ne!(Luv { l: 50.0, u: 100.0, v: 75.0 },
+                   Luv { l: 50.0, u: 100.0, v: 25.0 });
+        assert_ne!(LCh { l: 50.0, c: 100.0, h: 1.0 },
+                   LCh { l: 25.0, c: 100.0, h: 1.0 });
+        assert_ne!(LCh { l: 50.0, c: 100.0, h: 1.0 },
+                   LCh { l: 50.0, c:  50.0, h: 1.0 });
+        assert_ne!(LCh { l: 50.0, c: 100.0, h: 1.0 },
+                   LCh { l: 50.0, c: 100.0, h: 2.0 });
     }
 }
